import os
import sys
import time
import shutil
from subprocess import call, Popen
from annogesiclib.multiparser import Multiparser
from annogesiclib.helper import Helper
from annogesiclib.converter import Converter
from annogesiclib.circRNA import detect_circrna


class CircRNADetection(object):

    def __init__(self, args_circ):
        self.multiparser = Multiparser()
        self.helper = Helper()
        self.converter = Converter()
        self.alignment_path = os.path.join(args_circ.output_folder,
                                           "segemehl_align")
        self.splice_path = os.path.join(args_circ.output_folder,
                                        "segemehl_splice")
        self.candidate_path = os.path.join(args_circ.output_folder,
                                           "circRNA_tables")
        self.gff_folder = os.path.join(args_circ.output_folder, "gffs")
        self.gff_path = os.path.join(args_circ.gffs, "tmp")
        self.splices = {"all_file": "splicesites_all.bed",
                        "file": "splicesites.bed",
                        "all": "splicesites_all", "splice": "splicesites"}
        self.trans = {"all_file": "transrealigned_all.bed",
                      "file": "transrealigned.bed",
                      "all": "transrealigned_all", "trans": "transrealigned"}
        self.bams = {"whole": "whole_reads.bam", "sort": "whole_reads_sort"}
        if args_circ.align:
            if args_circ.fastas is None:
                print("Error: There is no genome fasta file!!!")
                sys.exit()
            else:
                self.fasta_path = os.path.join(args_circ.fastas, "tmp")
        else:
            self.fasta_path = os.path.join(args_circ.fastas, "tmp")

    def _wait_process(self, processes):
        for p in processes:
            p.wait()
            if p.stdout:
                p.stdout.close()
            if p.stdin:
                p.stdin.close()
            if p.stderr:
                p.stderr.close()
            try:
                p.kill()
            except OSError:
                pass
            time.sleep(5)

    def _deal_zip_file(self, read_folder):
        tmp_reads = []
        for read in os.listdir(read_folder):
            if read.endswith(".bz2"):
                mod_read = read.replace(".bz2", "")
                if (".fa" not in mod_read) and (
                        ".fasta" not in mod_read) and (
                        ".fna" not in mod_read):
                    mod_read = mod_read + ".fa"
                read_out = open(os.path.join(read_folder, mod_read), "w")
                tmp_reads.append(os.path.join(read_folder, mod_read))
                print(" ".join(["unzip", read]))
                call(["bzcat", os.path.join(read_folder, read)],
                     stdout=read_out)
                read_out.close()
            elif read.endswith(".gz"):
                mod_read = read.replace(".gz", "")
                if (".fa" not in mod_read) and (
                        ".fasta" not in mod_read) and (
                        ".fna" not in mod_read):
                    mod_read = mod_read + ".fa"
                read_out = open(os.path.join(read_folder, mod_read), "w")
                tmp_reads.append(os.path.join(read_folder, mod_read))
                print(" ".join(["unzip", read]))
                call(["zcat", os.path.join(read_folder, read)],
                     stdout=read_out)
                read_out.close()
        return tmp_reads

    def _run_segemehl_fasta_index(self, segemehl_path, fasta_path,
                                  index, fasta):
        call([os.path.join(segemehl_path, "segemehl.x"),
              "-x", os.path.join(fasta_path, index),
              "-d", os.path.join(fasta_path, fasta)])

    def _run_segemehl_align(self, args_circ, index, fasta, read,
                            sam_file, log_file, fasta_prefix):
        out = open(os.path.join(self.alignment_path,
                   fasta_prefix, sam_file), "w")
        log = open(os.path.join(self.alignment_path,
                   fasta_prefix, log_file), "w")
        p = Popen([os.path.join(args_circ.segemehl_path, "segemehl.x"),
                   "-i", os.path.join(self.fasta_path, index),
                   "-d", os.path.join(self.fasta_path, fasta),
                   "-q", os.path.join(args_circ.read_folder, read), "-S"],
                  stdout=out, stderr=log)
        return p

    def _align(self, args_circ):
        prefixs = []
        align_files = []
        for fasta in os.listdir(self.fasta_path):
            index = fasta.replace(".fa", ".idx")
            self._run_segemehl_fasta_index(args_circ.segemehl_path,
                                           self.fasta_path, index, fasta)
            processes = []
            num_process = 0
            fasta_prefix = fasta.replace(".fa", "")
            prefixs.append(fasta_prefix)
            self.helper.check_make_folder(os.path.join(
                                self.alignment_path, fasta_prefix))
            for read in os.listdir(args_circ.read_folder):
                num_process += 1
                if read.endswith(".fa") or \
                   read.endswith(".fna") or \
                   read.endswith("fasta"):
                    filename = read.split(".")
                    read_prefix = ".".join(filename[:-1])
                    sam_file = "_".join([read_prefix, fasta_prefix + ".sam"])
                    log_file = "_".join([read_prefix, fasta_prefix + ".log"])
                    align_files.append("_".join([read_prefix, fasta_prefix]))
                    print("mapping {0}".format(sam_file))
                    p = self._run_segemehl_align(
                            args_circ, index, fasta, read,
                            sam_file, log_file, fasta_prefix)
                    processes.append(p)
                    if num_process == args_circ.cores:
                        self._wait_process(processes)
                        num_process = 0
            self._wait_process(processes)
        return align_files, prefixs

    def _run_samtools_convert_bam(self, samtools_path, pre_sam, out_bam):
        call([samtools_path, "view", "-bS", pre_sam, "-o", out_bam])

    def _convert_sam2bam(self, sub_alignment_path, samtools_path, align_files):
        bam_files = []
        convert_ones = []
        remove_ones = []
        for sam in os.listdir(sub_alignment_path):
            pre_sam = os.path.join(sub_alignment_path, sam)
            if sam.endswith(".sam"):
                bam_file = sam.replace(".sam", ".bam")
                print("Convert {0} to {1}".format(sam, bam_file))
                out_bam = os.path.join(sub_alignment_path, bam_file)
                self._run_samtools_convert_bam(samtools_path, pre_sam, out_bam)
                bam_files.append(out_bam)
                if align_files:
                    if bam_file.replace(".bam", "") not in align_files:
                        convert_ones.append(out_bam)
                    else:
                        remove_ones.append(pre_sam)
            elif sam.endswith(".bam"):
                if (pre_sam not in convert_ones) and (
                        pre_sam not in remove_ones):
                    bam_files.append(pre_sam)
            elif sam.endswith(".log"):
                os.remove(pre_sam)
        return bam_files, convert_ones, remove_ones

    def _run_samtools_merge_sort(self, samtools_path,
                                 sub_alignment_path, bam_files):
        print("Merge all bam files....")
        whole_bam = os.path.join(sub_alignment_path, self.bams["whole"])
        if len(bam_files) <= 1:
            shutil.copyfile(bam_files[0], whole_bam)
        else:
            file_line = " ".join(bam_files)
            os.system(" ".join([samtools_path, "merge",
                                whole_bam, file_line]))
        print("Sort bam files....")
        call([samtools_path, "sort", "-o", os.path.join(sub_alignment_path,
              self.bams["sort"] + ".bam"), whole_bam])
        os.remove(os.path.join(sub_alignment_path, self.bams["whole"]))

    def _run_samtools_convert_sam(self, samtools_path, sub_alignment_path):
        print("Convert whole reads bam file to sam file....")
        call([samtools_path, "view", "-h", "-o",
              os.path.join(sub_alignment_path, self.bams["sort"] + ".sam"),
              os.path.join(sub_alignment_path, self.bams["sort"] + ".bam")])

    def _merge_sort_aligment_file(self, bam_files, samtools_path,
                                  sub_alignment_path, convert_ones,
                                  tmp_reads, remove_ones):
        self._run_samtools_merge_sort(samtools_path,
                                      sub_alignment_path, bam_files)
        self._run_samtools_convert_sam(samtools_path, sub_alignment_path)
        for bam in convert_ones:
            os.remove(bam)
        for sam in remove_ones:
            os.remove(sam)
        if len(tmp_reads) != 0:
            for read in tmp_reads:
                os.remove(read)

    def _run_testrealign(self, prefix, segemehl_path, sub_alignment_path):
        self.helper.check_make_folder(os.path.join(self.splice_path, prefix))
        sub_splice_path = os.path.join(self.splice_path, prefix)
        err_log = os.path.join(sub_splice_path, prefix + ".log")
        print("Running testrealign.x for {0}".format(prefix))
        command = " ".join([
                  os.path.join(segemehl_path, "testrealign.x"),
                  "-d", os.path.join(self.fasta_path, prefix + ".fa"),
                  "-q", os.path.join(sub_alignment_path,
                                     self.bams["sort"] + ".sam"),
                  "-n"])
        os.system(command + " 2>" + err_log)
        self.helper.move_all_content(os.getcwd(), sub_splice_path, [".bed"])
        self.helper.remove_all_content(sub_alignment_path,
                                       self.bams["sort"], "file")

    def _merge_bed(self, fastas, splice_path):
        tmp_prefixs = []
        for fasta in os.listdir(fastas):
            headers = []
            if (fasta.endswith(".fa") or fasta.endswith(".fna") or
                    fasta.endswith(".fasta")):
                with open(os.path.join(fastas, fasta), "r") as f_h:
                    for line in f_h:
                        line = line.strip()
                        if line.startswith(">"):
                            headers.append(line[1:])
                filename = fasta.split(".")
                fasta_prefix = ".".join(filename[:-1])
                tmp_prefixs.append(fasta_prefix)
                self.helper.check_make_folder(os.path.join(
                                              os.getcwd(), fasta_prefix))
                for header in headers:
                    shutil.copyfile(os.path.join(splice_path, header,
                                    self.splices["file"]),
                                    os.path.join(fasta_prefix,
                                    "_".join([self.splices["splice"],
                                              header + ".bed"])))
                    shutil.copyfile(os.path.join(splice_path, header,
                                    self.trans["file"]),
                                    os.path.join(fasta_prefix,
                                    "_".join([self.trans["trans"],
                                              header + ".bed"])))
                out_splice = os.path.join(fasta_prefix,
                                          self.splices["all_file"])
                out_trans = os.path.join(fasta_prefix,
                                         self.trans["all_file"])
                if len(headers) > 1:
                    for file_ in os.listdir(fasta_prefix):
                        if (self.splices["splice"] in file_) and (
                                self.splices["all"] not in file_):
                            self.helper.merge_file(os.path.join(
                                    fasta_prefix, file_), out_splice)
                        elif (self.trans["trans"] in file_) and (
                                self.trans["all"] not in file_):
                            self.helper.merge_file(os.path.join(
                                    fasta_prefix, file_), out_trans)
                else:
                    shutil.move(os.path.join(
                                fasta_prefix,
                                "_".join([self.splices["splice"],
                                         headers[0] + ".bed"])),
                                out_splice)
                    shutil.move(os.path.join(
                                fasta_prefix,
                                "_".join([self.trans["trans"],
                                          headers[0] + ".bed"])),
                                out_trans)
        self.helper.remove_all_content(splice_path, None, "dir")
        return tmp_prefixs

    def _stat_and_gen_gff(self, tmp_prefixs, args_circ):
        for prefix in tmp_prefixs:
            self.helper.check_make_folder(os.path.join(self.gff_folder,
                                                       prefix))
            shutil.copytree(prefix, os.path.join(self.splice_path, prefix))
            self.helper.check_make_folder(os.path.join(
                                          self.candidate_path, prefix))
            print("comparing with annotation of {0}".format(prefix))
            if self.splices["all_file"] in os.listdir(os.path.join(
                                           self.splice_path, prefix)):
                detect_circrna(os.path.join(self.splice_path, prefix,
                               self.splices["all_file"]), os.path.join(
                               self.gff_path, prefix + ".gff"),
                               os.path.join(self.candidate_path, prefix,
                               "_".join(["circRNA", prefix + "_all.csv"])),
                               args_circ, os.path.join(args_circ.stat_folder,
                               "_".join(["stat_circRNA", prefix + ".csv"])))
                self.converter.convert_circ2gff(
                     os.path.join(self.candidate_path, prefix,
                                  "_".join(["circRNA",
                                            prefix + "_all.csv"])),
                     args_circ, os.path.join(
                                self.gff_folder, prefix,
                                "_".join([prefix, "circRNA_all.gff"])),
                     os.path.join(self.gff_folder, prefix,
                                  "_".join([prefix, "circRNA_best.gff"])))

    def _assign_merge_bam(self, args_circ):
        remove_frags = []
        bam_files = []
        if (args_circ.normal_bams is not None) and (
                args_circ.frag_bams is not None):
            for frag in os.listdir(args_circ.frag_bams):
                if frag.endswith(".bam"):
                    shutil.copyfile(os.path.join(args_circ.frag_bams, frag),
                                    os.path.join(args_circ.normal_bams, frag))
                    remove_frags.append(frag)
            merge_folder = args_circ.normal_bams
        elif (args_circ.normal_bams is not None):
            merge_folder = args_circ.normal_bams
        elif (args_circ.frag_bams is not None):
            merge_folder = args_circ.frag_bams
        else:
            print("Error: please assign bam folder or do alignment!!")
            sys.exit()
        for bam in os.listdir(merge_folder):
            if bam.endswith(".bam"):
                bam_files.append(os.path.join(merge_folder, bam))
        return merge_folder, remove_frags, bam_files

    def run_circrna(self, args_circ):
        for gff in os.listdir(args_circ.gffs):
            if gff.endswith(".gff"):
                self.helper.check_uni_attributes(os.path.join(
                                                 args_circ.gffs, gff))
        if args_circ.segemehl_path is None:
            print("Error: please assign segemehl folder!!")
            sys.exit()
        self.multiparser.parser_gff(args_circ.gffs, None)
        self.multiparser.combine_gff(args_circ.fastas, self.gff_path,
                                     "fasta", None)
        tmp_reads = []
        if args_circ.align:
            self.multiparser.parser_fasta(args_circ.fastas)
            tmp_reads = self._deal_zip_file(args_circ.read_folder)
            align_files, prefixs = self._align(args_circ)
        else:
            self.multiparser.parser_fasta(args_circ.fastas)
            prefixs = []
            for fasta in os.listdir(self.fasta_path):
                fasta_prefix = fasta.replace(".fa", "")
                prefixs.append(fasta_prefix)
            merge_folder, remove_frag, bam_files = self._assign_merge_bam(
                                                   args_circ)
            align_files = None
        for prefix in prefixs:
            if args_circ.align:
                sub_alignment_path = os.path.join(self.alignment_path, prefix)
                bam_files, convert_ones, remove_ones = self._convert_sam2bam(
                    sub_alignment_path, args_circ.samtools_path, align_files)
            else:
                sub_alignment_path = merge_folder
                convert_ones = []
                remove_ones = []
            self._merge_sort_aligment_file(
                bam_files, args_circ.samtools_path, sub_alignment_path,
                convert_ones, tmp_reads, remove_ones)
            self._run_testrealign(prefix, args_circ.segemehl_path,
                                  sub_alignment_path)
        tmp_prefixs = self._merge_bed(args_circ.fastas, self.splice_path)
        self.multiparser.parser_gff(args_circ.gffs, None)
        self.multiparser.combine_gff(args_circ.fastas, self.gff_path,
                                     "fasta", None)
        self._stat_and_gen_gff(tmp_prefixs, args_circ)
        self.helper.remove_tmp(args_circ.fastas)
        self.helper.remove_tmp(args_circ.gffs)
        for tmp_prefix in tmp_prefixs:
            shutil.rmtree(tmp_prefix)
        if (not args_circ.align) and (len(remove_frag) != 0):
            for frag in remove_frag:
                os.remove(os.path.join(merge_folder, frag))



import os
import sys
from annogesiclib.parser_wig import WigParser


def check_tex_conds(tracks, libs, texs, check_tex, conds, tex_notex):
    for track in tracks:
        for lib in libs:
            if lib["name"] == track:
                if "tex" in lib["type"]:
                    type_ = "tex"
                else:
                    type_ = "frag"
                index = "_".join([lib["cond"], type_])
                if len(texs) != 0:
                    for key, num in texs.items():
                        if track in key:
                            if (texs[key] >= tex_notex) and (
                                    key not in check_tex):
                                check_tex.append(key)
                                if index not in conds.keys():
                                    conds[index] = 1
                                else:
                                    conds[index] += 1
                else:
                    if index not in conds.keys():
                        conds[index] = 1
                    else:
                        conds[index] += 1


def detect_hight_toler(cover, height, tmp_covers, tracks):
    if cover["coverage"] > height:
        if tmp_covers["best"] < cover["coverage"]:
            tmp_covers["best"] = cover["coverage"]
        tracks.append(cover["track"])
    else:
        if cover["coverage"] > tmp_covers["toler"]:
            tmp_covers["toler"] = cover["coverage"]


def elongation(covers, template_texs, libs, strand, trans,
               args_tran, strain, tolers):
    first = True
    pre_pos = -1
    check_tex = []
    tracks = []
    conds = {}
    pre_wig = None
    texs = template_texs.copy()
    tmp_covers = {"best": 0, "toler": -1}
    for cover in covers:
        if (pre_pos == cover["pos"]) or first:
            first = False
            detect_hight_toler(cover, args_tran.height, tmp_covers, tracks)
        else:
            for track in tracks:
                if len(texs) != 0:
                    for key, num in texs.items():
                        if track in key:
                            texs[key] += 1
            check_tex_conds(tracks, libs, texs, check_tex,
                            conds, args_tran.tex)
            for cond, num in conds.items():
                if ((num >= args_tran.replicates["tex"]) and (
                     "tex" in cond)) or (
                    (num >= args_tran.replicates["frag"]) and (
                     "frag" in cond)):
                    trans[strain].append({
                        "strand": strand, "pos": pre_wig["pos"],
                        "coverage": tmp_covers["best"], "cond": num})
            if (tmp_covers["toler"] != -1):
                tolers.append(tmp_covers["toler"])
            else:
                tolers.append(args_tran.height + 10)
            tmp_covers = {"best": 0, "toler": -1}
            tracks = []
            conds = {}
            check_tex = []
            texs = template_texs.copy()
            detect_hight_toler(cover, args_tran.height, tmp_covers, tracks)
        pre_wig = cover
        pre_pos = cover["pos"]
    return tmp_covers["best"], conds, tracks, texs, pre_pos


def transfer_to_tran(wigs, libs, template_texs, strand, args_tran):
    tolers = {}
    trans = {}
    for strain, covers in wigs.items():
        if strain not in trans:
            trans[strain] = []
            tolers[strain] = []
        sort_covers = sorted(covers, key=lambda k: (k["pos"], k["cond"]))
        best_cover, conds, tracks, texs, pos = elongation(
                sort_covers, template_texs, libs, strand, trans,
                args_tran, strain, tolers[strain])
    for track in tracks:
        if len(texs) != 0:
            for key, num in texs.items():
                if track in key:
                    texs[key] += 1
    for track in tracks:
        for lib in libs:
            if lib["name"] == track:
                index = "_".join([lib["cond"], lib["type"]])
                if len(texs) != 0:
                    for key, num in texs.items():
                        if (texs[key] >= args_tran.tex) and (
                                lib["name"] in key):
                            if index not in conds.keys():
                                conds[index] = 1
                            else:
                                conds[index] += 1
                else:
                    if index not in conds.keys():
                        conds[index] = 1
                    else:
                        conds[index] += 1
    for cond, num in conds.items():
        if ((num >= args_tran.replicates["tex"]) and ("tex" in cond)) or (
                (num >= args_tran.replicates["frag"]) and ("frag" in cond)):
            trans[strain].append({"strand": strand, "pos": pos,
                                  "coverage": best_cover, "cond": num})
    return tolers, trans


def print_transctipt(start, end, width, num, high_cover, wig_type,
                     low_cover, out, strain, strand):
    if (start != -1) and (end != -1) and (
            (end - start) >= width):
        name = '%0*d' % (5, num)
        attribute = gen_attribute_string(num, name, high_cover,
                                         low_cover, wig_type)
        out.write("\t".join([str(field) for field in [
                  strain, "ANNOgesic", "transcript", str(start),
                  str(end), ".", strand, ".", attribute]]) + "\n")


def gen_attribute_string(num, name, high_cover, low_cover, wig_type):
    attribute = ";".join(
                ["=".join(items) for items in ([
                    ("ID", "tran_" + str(num)),
                    ("Name", "transcript_" + name),
                    ("high_coverage", str(high_cover)),
                    ("low_coverage", str(low_cover)),
                    ("detect_lib", wig_type)])])
    return attribute


def fill_gap_and_print(trans, strand, out, tolers, wig_type, args_tran):
    for strain, datas in tolers.items():
        num = 0
        for data in datas:
            num += 1
    for strain, covers in trans.items():
        covers = sorted(covers, key=lambda k: k["pos"])
        first = True
        start = -1
        end = -1
        num = 0
        pre_cover = None
        for cover in covers:
            fit = True
            if first:
                first = False
                start = cover["pos"]
                high_cover = cover["coverage"]
                low_cover = cover["coverage"]
            else:
                if (cover["pos"] - pre_cover["pos"]) <= args_tran.tolerance:
                    if cover["pos"] - pre_cover["pos"] > 1:
                        for toler_strain, toler_datas in tolers.items():
                            if toler_strain == strain:
                                toler_covers = toler_datas[
                                        (pre_cover["pos"] - 1): cover["pos"]]
                                for toler_cover in toler_covers:
                                    if (toler_cover < args_tran.low_cutoff):
                                        fit = False
                                        break
                    if fit:
                        end = cover["pos"]
                        if high_cover < cover["coverage"]:
                            high_cover = cover["coverage"]
                        if low_cover > cover["coverage"]:
                            low_cover = cover["coverage"]
                if ((cover["pos"] - pre_cover["pos"]) >
                        args_tran.tolerance) or (not fit):
                    if (start != -1) and (end != -1) and (
                            (end - start) >= args_tran.width):
                        name = '%0*d' % (5, num)
                        attribute = gen_attribute_string(num, name, high_cover,
                                                         low_cover, wig_type)
                        out.write("\t".join([str(field) for field in [
                            strain, "ANNOgesic", "transcript", str(start),
                            str(end), ".", strand, ".", attribute]]) + "\n")
                        num += 1
                    start = cover["pos"]
                    end = -1
                    high_cover = cover["coverage"]
                    low_cover = cover["coverage"]
            pre_cover = cover
        if len(covers) != 0:
            print_transctipt(start, end, args_tran.width, num, high_cover,
                             wig_type, low_cover, out, strain, strand)


def read_wig(filename, libs, strand):
    wigs = {}
    tracks = []
    wig_parser = WigParser()
    wig_fh = open(filename)
    for entry in wig_parser.parser(wig_fh, strand):
        if entry.track not in tracks:
            tracks.append(entry.track)
        if entry.strain not in wigs.keys():
            strain = entry.strain
            wigs[entry.strain] = []
        for lib in libs:
            if strand == lib["strand"]:
                for key, value in lib.items():
                    if (key == "name") and (value == entry.track):
                        cond = lib["cond"]
                        track = value
                        wigs[strain].append({
                              "pos": entry.pos, "coverage": entry.coverage,
                              "strand": entry.strand, "track": track,
                              "cond": cond})
                        break
    wig_fh.close()
    return wigs


def assembly(wig_f_file, wig_r_file, wig_folder, input_lib,
             out_file, wig_type, args_tran):
    libs = []
    conds = []
    out = open(out_file, "w")
    out.write("##gff-version 3\n")
    for lib in input_lib:
        datas = lib.split(":")
        if datas[2] not in conds:
            conds.append(datas[2])
        for wig in os.listdir(wig_folder):
            if wig == datas[0]:
                with open(wig_folder + "/" + wig, "r") as w_h:
                    for line in w_h:
                        line = line.strip()
                        if line.startswith("track"):
                            name = line.split("=")[-1][1:-1]
                            break
        libs.append({"name": name, "type": datas[1],
                     "cond": datas[2], "rep": datas[3], "strand": datas[4]})
    texs = {}
    for lib1 in libs:
        if lib1["type"] == "frag":
            pass
        elif (lib1["type"] == "tex") or (lib1["type"] == "notex"):
            for lib2 in libs:
                if (lib1["cond"] == lib2["cond"]) and (
                        lib1["rep"] == lib2["rep"]) and (
                        lib1["type"] == "tex") and (
                        lib2["type"] == "notex") and (
                        lib1["strand"] == lib2["strand"]):
                    texs[lib1["name"] + "_" + lib2["name"]] = 0
        else:
            print("Error:not correct library type "
                  "(fragmented or Tex treated??)")
            sys.exit()
    wig_fs = read_wig(wig_f_file, libs, "+")
    wig_rs = read_wig(wig_r_file, libs, "-")
    tolers_f, tran_fs = transfer_to_tran(wig_fs, libs, texs, "+", args_tran)
    tolers_r, tran_rs = transfer_to_tran(wig_rs, libs, texs, "-", args_tran)
    fill_gap_and_print(tran_fs, "+", out, tolers_f, wig_type, args_tran)
    fill_gap_and_print(tran_rs, "-", out, tolers_r, wig_type, args_tran)
    out.close()



import os
import shutil
from annogesiclib.multiparser import Multiparser
from annogesiclib.seq_editer import SeqEditer
from annogesiclib.helper import Helper


class TargetFasta(object):

    def __init__(self, tar_folder, ref_folder):
        self.multiparser = Multiparser()
        self.seq_editer = SeqEditer()
        self.helper = Helper()
        self.folders = {"tmp_tar": os.path.join(tar_folder, "tmp"),
                        "tmp_ref": os.path.join(ref_folder, "tmp")}

    def get_target_fasta(self, mut_table, tar_folder, ref_folder, output):
        self.multiparser.parser_fasta(ref_folder)
        if "tmp" in os.listdir(tar_folder):
            shutil.rmtree(self.folders["tmp_tar"])
        os.mkdir(self.folders["tmp_tar"])
        self.seq_editer.modify_seq(self.folders["tmp_ref"], mut_table,
                                   self.folders["tmp_tar"])
        print("transfer to target fasta...")
        if output is not None:
            for file_ in output:
                first = True
                datas = file_.split(":")
                filename = datas[0]
                strains = datas[1].split("_and_")
                out = open(os.path.join(tar_folder, filename + ".fa"), "w")
                for strain in strains:
                    if strain + ".fa" in os.listdir(self.folders["tmp_tar"]):
                        if first:
                            first = False
                        else:
                            out.write("\n")
                        with open(os.path.join(
                                  self.folders["tmp_tar"],
                                  strain + ".fa")) as f_h:
                            for line in f_h:
                                out.write(line)
                    else:
                        print("Error:no fasta information of {0}.fa".format(
                              strain))
                out.close()
        else:
            self.helper.move_all_content(self.folders["tmp_tar"],
                                         tar_folder, [".fa"])
        shutil.rmtree(self.folders["tmp_tar"])
        shutil.rmtree(self.folders["tmp_ref"])
        self.helper.remove_all_content(ref_folder, "_folder", "dir")
        print("please use the new fasta file to remapping again.")
        print("Then copy BAMs and wigs back to input/align_results/BAMs "
              "and input/align_results/wigs")



import os
from annogesiclib.helper import Helper
from annogesiclib.gff3 import Gff3Parser
from annogesiclib.parser_wig import WigParser


def read_gff(gff_file, features):
    gffs = []
    if not os.path.isfile(gff_file):
        filename = gff_file.split(".")
        gff_file = ".".join(filename[0:-2]) + ".gff"
    g_f = open(gff_file, "r")
    for entry in Gff3Parser().entries(g_f):
        if entry.feature in features:
            gffs.append(entry)
    gffs = sorted(gffs, key=lambda k: (k.seq_id, k.start))
    return gffs


def is_primary(cds_start, cds_end, tss_pos, strand):
    if strand == "+":
        if is_utr(cds_start, tss_pos, 300) and (cds_start >= tss_pos):
            return True
    else:
        if is_utr(tss_pos, cds_end, 300) and (cds_end <= tss_pos):
            return True


def is_internal(cds_start, cds_end, tss_pos, strand):
    if ((cds_start < tss_pos) and (cds_end > tss_pos)) or (
            (strand == "+") and (tss_pos == cds_end)) or (
            (strand == "-") and (tss_pos == cds_start)):
        return True


def is_antisense(cds_start, cds_end, tss_pos, strand):
    if ((is_utr(cds_start, tss_pos, 100)) and (cds_start >= tss_pos)) or (
            (is_utr(tss_pos, cds_end, 100)) and (cds_end <= tss_pos)) or (
            is_internal(cds_start, cds_end, tss_pos, strand)):
        return True


def is_utr(pos1, pos2, length):
    if pos1 - pos2 <= length:
        return True


def get_attributes(tss, cds):
    if tss.attributes["associated_gene"] == "orphan":
        if "locus_tag" in cds.attributes.keys():
            tss.attributes["associated_gene"] = cds.attributes["locus_tag"]
        else:
            strand = Helper().get_strand_name(cds.strand)
            tss.attributes["associated_gene"] = cds.feature + ":" + \
                str(cds.start) + "-" + str(cds.end) + "_" + strand
    else:
        if "locus_tag" in cds.attributes.keys():
            tss.attributes["associated_gene"] = "&".join([
                tss.attributes["associated_gene"],
                cds.attributes["locus_tag"]])
        else:
            strand = Helper().get_strand_name(cds.strand)
            tss.attributes["associated_gene"] = "&".join([
                tss.attributes["associated_gene"],
                cds.feature + ":" + str(cds.start) + "-" +
                str(cds.end) + "_" + strand])


def detect_coverage(wigs, tss, ref):
    tss_cover = -1
    ref_cover = -1
    for strain, tracks in wigs.items():
        if strain == tss.seq_id:
            tss_cover = 0
            ref_cover = 0
            for wig in tracks.values():
                if ((tss.start + 1) <= len(wig)) and (
                        (ref.start + 1) <= len(wig)):
                    if tss.strand == "+":
                        diff_t = (wig[tss.start - 1]["coverage"] -
                                  wig[tss.start - 2]["coverage"])
                        diff_r = (wig[ref.start - 1]["coverage"] -
                                  wig[ref.start - 2]["coverage"])
                    else:
                        diff_t = (wig[tss.start - 1]["coverage"] -
                                  wig[tss.start]["coverage"])
                        diff_r = (wig[ref.start - 1]["coverage"] -
                                  wig[ref.start]["coverage"])
                    tss_cover = tss_cover + diff_t
                    ref_cover = ref_cover + diff_r
    return tss_cover, ref_cover


def del_repeat(tsss):
    for tss in tsss:
        types = tss.attributes["type"].split("&")
        utrs = tss.attributes["UTR_length"].split("&")
        genes = tss.attributes["associated_gene"].split("&")
        detect_pri = False
        detect_sec = False
        index = 0
        final_types = []
        final_utrs = []
        final_genes = []
        for type_ in types:
            if (type_ == "Primary") and (not detect_pri):
                detect_pri = True
                pri_utr = int(utrs[index].split("_")[1])
                real_index = index
            elif (type_ == "Primary") and (detect_pri):
                compare_utr = int(utrs[index].split("_")[1])
                if compare_utr < pri_utr:
                    pri_utr = compare_utr
                    real_index = index
            elif (type_ == "Secondary") and (not detect_sec):
                detect_sec = True
                sec_utr = int(utrs[index].split("_")[1])
                real_index2 = index
            elif (type_ == "Secondary") and (detect_sec):
                compare_utr = int(utrs[index].split("_")[1])
                if compare_utr < sec_utr:
                    sec_utr = compare_utr
                    real_index2 = index
            elif (type_ == "Antisense") or (type_ == "Internal") or (
                    type_ == "Orphan"):
                final_types.append(types[index])
                final_utrs.append(utrs[index])
                final_genes.append(genes[index])
            index += 1
        if detect_pri:
            final_types.append(types[real_index])
            final_utrs.append(utrs[real_index])
            final_genes.append(genes[real_index])
        else:
            if detect_sec:
                final_types.append(types[real_index2])
                final_utrs.append(utrs[real_index2])
                final_genes.append(genes[real_index2])
        tss.attributes["type"] = "&".join(final_types)
        tss.attributes["UTR_length"] = "&".join(final_utrs)
        tss.attributes["associated_gene"] = "&".join(final_genes)


def fix_attributes(tss, tss_entry):
    index = 0
    genes = tss.attributes["associated_gene"].split("&")
    utrs = tss.attributes["UTR_length"].split("&")
    types = tss.attributes["type"].split("&")
    for gene in genes:
        if gene == tss_entry["locus"]:
            utrs[index] = utrs[index].replace("Primary", "Secondary")
            types[index] = types[index].replace("Primary", "Secondary")
        index += 1
    tss.attributes["UTR_length"] = "&".join(utrs)
    tss.attributes["type"] = "&".join(types)


def get_primary_locus_tag(tss):
    tsss = []
    tss_types = tss.attributes["type"].split("&")
    tss_locus_tags = tss.attributes["associated_gene"].split("&")
    tss_utr_lengths = tss.attributes["UTR_length"].split("&")
    index = 0
    for tss_type in tss_types:
        if "Primary" in tss_type:
            tsss.append({"locus": tss_locus_tags[index],
                         "utr": int(tss_utr_lengths[index].split("_")[1]),
                         "type": tss_type})
        index += 1
    return tsss


def fix_primary_type(tsss, wigs_f, wigs_r):
    for tss in tsss:
        if "Primary" in tss.attributes["type"]:
            tss_entrys = get_primary_locus_tag(tss)
            for ref in tsss:
                if (ref.seq_id == tss.seq_id) and (
                    ref.strand == tss.strand) and (
                        ref.start == tss.start):
                    pass
                else:
                    if "Primary" in ref.attributes["type"]:
                        ref_entrys = get_primary_locus_tag(ref)
                        for tss_entry in tss_entrys:
                            for ref_entry in ref_entrys:
                                if (tss_entry["locus"] ==
                                        ref_entry["locus"]) and (
                                        tss_entry["type"] == "Primary") and (
                                        ref_entry["type"] == "Primary") and (
                                        tss.seq_id == ref.seq_id):
                                    if tss.strand == "+":
                                        tss_cover, ref_cover = detect_coverage(
                                            wigs_f, tss, ref)
                                    else:
                                        tss_cover, ref_cover = detect_coverage(
                                            wigs_r, tss, ref)
                                    if tss_cover < ref_cover:
                                        fix_attributes(tss, tss_entry)
                                    elif tss_cover > ref_cover:
                                        fix_attributes(ref, ref_entry)
                                    elif tss_cover == ref_cover:
                                        if tss_entry["utr"] < ref_entry["utr"]:
                                            fix_attributes(ref, ref_entry)
                                        elif (tss_entry["utr"] >
                                              ref_entry["utr"]):
                                            fix_attributes(tss, tss_entry)
    del_repeat(tsss)
    return tsss


def read_wig(filename, strand):
    wigs = {}
    wig_parser = WigParser()
    if filename:
        wig_fh = open(filename)
        for entry in wig_parser.parser(wig_fh, strand):
            if entry.strain not in wigs.keys():
                strain = entry.strain
                wigs[strain] = {}
            if entry.track not in wigs[strain].keys():
                wigs[strain][entry.track] = []
            wigs[strain][entry.track].append({
                 "pos": entry.pos, "coverage": entry.coverage,
                 "strand": entry.strand})
        wig_fh.close()
    return wigs


def get_attributes_int_anti(tss, cds, type_):
    if tss.attributes["type"] != "Orphan":
        tss.attributes["type"] = "&".join(
                [tss.attributes["type"], type_])
        tss.attributes["UTR_length"] = "&".join(
                [tss.attributes["UTR_length"],
                 type_ + "_NA"])
    else:
        tss.attributes["type"] = type_
        tss.attributes["UTR_length"] = type_ + "_NA"
    get_attributes(tss, cds)


def compare_cds_check_orphan(tsss, cdss):
    for tss in tsss:
        if tss.attributes["type"] == "Orphan":
            for cds in cdss:
                if (tss.seq_id == cds.seq_id) and \
                   (tss.strand == cds.strand):
                    if is_primary(cds.start, cds.end, tss.start, tss.strand):
                        if tss.attributes["type"] != "Orphan":
                            tss.attributes["type"] = "&".join(
                                    [tss.attributes["type"], "Primary"])
                            if tss.strand == "+":
                                tss.attributes["UTR_length"] = "&".join([
                                    tss.attributes["UTR_length"],
                                    "Primary_" + str(cds.start - tss.start)])
                            else:
                                tss.attributes["UTR_length"] = "&".join([
                                    tss.attributes["UTR_length"],
                                    "Primary_" + str(tss.start - cds.end)])
                        else:
                            tss.attributes["type"] = "Primary"
                            if tss.strand == "+":
                                tss.attributes["UTR_length"] = (
                                    "Primary_" + str(cds.start - tss.start))
                            else:
                                tss.attributes["UTR_length"] = (
                                    "Primary_" + str(tss.start - cds.end))
                        get_attributes(tss, cds)
                    if is_internal(cds.start, cds.end, tss.start, tss.strand):
                        if "locus_tag" in cds.attributes.keys():
                            if (cds.attributes["locus_tag"] not in
                                    tss.attributes["associated_gene"]):
                                get_attributes_int_anti(tss, cds, "Internal")
                        else:
                            strand = Helper().get_strand_name(cds.strand)
                            if ("".join([cds.feature, ":", str(cds.start),
                                "-", str(cds.end), "_", strand]) not in
                                    tss.attributes["associated_gene"]):
                                get_attributes_int_anti(tss, cds, "Internal")
                    if is_antisense(cds.start, cds.end, tss.start, tss.strand):
                        if "locus_tag" in cds.attributes.keys():
                            if (cds.attributes["locus_tag"] not in
                                    tss.attributes["associated_gene"]):
                                get_attributes_int_anti(tss, cds, "Antisense")
                        else:
                            strand = Helper().get_strand_name(cds.strand)
                            if ("".join([cds.feature, ":", str(cds.start),
                                "-", str(cds.end), "_", strand]) not in
                                    tss.attributes["associated_gene"]):
                                get_attributes_int_anti(tss, cds, "Antisense")


def check_orphan(tss_file, gff_file, wig_f_file, wig_r_file, out_gff):
    cdss = read_gff(gff_file, ["CDS", "tRNA", "rRNA"])
    tsss = read_gff(tss_file, ["TSS"])
    wigs_f = read_wig(wig_f_file, "+")
    wigs_r = read_wig(wig_r_file, "-")
    out = open(out_gff, "w")
    out.write("##gff-version 3\n")
    compare_cds_check_orphan(tsss, cdss)
    final_tsss = fix_primary_type(tsss, wigs_f, wigs_r)
    for tss in final_tsss:
        tss.attribute_string = ";".join(
            ["=".join(items) for items in tss.attributes.items()])
        out.write("\t".join([str(field) for field in [
                        tss.seq_id, tss.source, tss.feature, tss.start,
                        tss.end, tss.score, tss.strand, tss.phase,
                        tss.attribute_string]]) + "\n")



from annogesiclib.gff3 import Gff3Parser


def gen_promoter_table(input_file, output_file, tss_file):
    tsss = []
    gff_f = open(tss_file, "r")
    for entry in Gff3Parser().entries(gff_f):
        tsss.append(entry)
    out = open(output_file, "w")
    out.write("\t".join(["strain", "TSS_position",
                         "TSS_strand", "Motif"]) + "\n")
    detect = False
    with open(input_file) as fh:
        for line in fh:
            line = line.strip()
            if line.startswith("MOTIF"):
                motif = line.split("MEME")[0].strip()
                datas = motif.split(" ")
                motif = datas[0] + "_" + datas[-1]
                detect = False
            elif (line.startswith("Sequence name")) and (
                    line.endswith("Site")):
                detect = True
            elif (len(line) == 0):
                detect = False
            elif (detect) and (not line.startswith("---")):
                tag = line.split(" ")[0]
                datas = tag.split("_")
                for tss in tsss:
                    if ("_".join(datas[2:]) in tss.seq_id) and (
                            datas[0] == str(tss.start)) and (
                            datas[1] == tss.strand):
                        out.write("\t".join([tss.seq_id, datas[0],
                                             datas[1], motif]) + "\n")
